// The bakery algorithm for mutual exclusion simulates a bakery in
// which customers (processes) take a numbered ticket when they enter the
// store, and then wait for their number to be called.
//
// This implementation has two processes, whose interleaving is
// emulated in the system process by permit variables.  Ties among
// processes with the same ticket number are broken according to a
// fixed priority scheme:  a process with lower index has precedence
// over one with higher index.
//
// The tickets are finite; hence, mutual exclusion is not guaranteed by
// this implementation.

`define N 2
// LSB of the ticket numbers.
`define LSB 1

// Type of program counter locations.
typedef enum { L1, L2, L3, L4, L5, L6, L7, L8, L9, L10, L11 } loc;

// The operating system (single processor).
module system (clk);
    input         clk;

    // More than one process may be choosing a ticket.  Hence, more than
    // one process may end up with the same ticket number.  These global
    // variables allow one process to wait for completion of a choice
    // that is in progress before comparing its ticket to that of another
    // process.  If the ticket number is the same, the process index is
    // used to decide which process accesses the critical section.
    reg           choosing0, choosing1;
    // The ticket numbers of the two processes.
    reg [0:`LSB]  number0, number1;
    // The next state values of the ticket numbers.
    wire [0:`LSB] num_out0, num_out1;
    // The next state values of the choosing variables.
    wire 	  ch_out0, ch_out1;
    // Process enabling variables.  Their mutual exclusion is ensured by
    // the scheduler.
    wire 	  permit0, permit1;
    // The latched values of the permit variables.  Unused by the system,
    // these variables appear in the fairness constraints.
    reg 	  st_perm0, st_perm1;

    // Scheduler: The enabled process is chosen nondeterministically.
    assign 	  permit0 = $ND(0, 1);
    assign 	  permit1 = !permit0;

    process p0(clk, permit0, choosing0, choosing1, ch_out0,
               number0, number1, num_out0, 0);
    process p1(clk, permit1, choosing0, choosing1, ch_out1,
               number0, number1, num_out1, 1);

    // Only the OS can update system memory.
    initial begin number0=0; number1=0; choosing0=0; choosing1=0; end
    initial st_perm0 = $ND(0,1);
    initial st_perm1 = $ND(0,1);
    always @ (posedge clk) begin
	st_perm0 = permit0; st_perm1 = permit1;
	if (permit0==1) begin number0=num_out0; choosing0=ch_out0; end
	if (permit1==1) begin number1=num_out1; choosing1=ch_out1; end
    end

endmodule // system


// Individual process.
module process (clk, permit, choosing0, choosing1, choosing_out,
                number0, number1, number_out, i);
    input           clk;
    input 	    permit;
    input 	    choosing0, choosing1;
    output 	    choosing_out;
    input [0:`LSB]  number0, number1;
    output [0:`LSB] number_out;
    input [0:1]     i;

    loc reg         pc;
    reg [0:1] 	    j;
    wire [0:`LSB]   numberi, numberj;
    wire 	    choosingj;
    wire 	    nond_exit;

    // The process may change its ticket number at two locations:
    //   L10: On exit from the critical section it takes number 0.
    //   L2 : When it wants to enter the critical section, the process
    //        takes the successor of the largest ticket numbers owned by
    //        the other processes.
    assign number_out = (pc==L10) ? 0 :
                 (pc==L2) ? ((number0 > number1) ? number0+1 : number1+1) :
                 (i==0) ? number0 : number1;
    // The  process starts choosing at location L1 and is done at location L3.
    assign choosing_out = (pc==L1) ? 1 : (pc==L3) ? 0 :
                 (i==0) ? choosing0 : choosing1;

    // These assignments emulate addressing into vectors.
    assign numberi = (i==0) ? number0 : number1;
    assign numberj = (j==0) ? number0 : number1;
    assign choosingj = (j==0) ? choosing0 : choosing1;

    // Nondeterministic choice of whether to exit the current location.
    assign nond_exit = $ND(0, 1);

    // The loop counter j could be nondeterministically initialized to
    // any value because it is set to 0 on entry to the loop anyway.
    initial begin pc=L1; j=0; end

    always @ (posedge clk) begin
	if (permit) begin
            case (pc)
	      // Declare the intention to pick a ticket.
	      // (See assignment to choosing_out.)
              L1: begin pc=L2; end
	      // Pick a ticket.  (See assignment to number_out.)
              L2: begin pc=L3; end
	      // Done with choosing.  (See assignment to choosing_out.)
              L3: begin pc=L4;end
	      // Loop over all processes to check ticket.
              L4: begin j=0; pc=L5; end
              L5: begin if (j<`N) pc=L6; else pc=L9; end
	      // Wait while choosing[j].
              L6: begin if (choosingj==1) pc=L6; else pc=L7; end
	      // Wait while process j has a non-zero ticket with a lower
	      // number, or it has the same number and higher priority.
              L7: begin
                  if ((numberj!=0) &&
                      (numberj<numberi || (numberj==numberi && j<i)))
                    pc=L7;
                  else
                    pc=L8;
              end
              L8: begin j=j+1; pc=L5; end
              // Enter critical section.
              L9: begin if (nond_exit==1) pc=L10; else pc=L9; end
	      // Leave critical section.
              L10: pc=L11;
              L11: begin if (nond_exit==1) pc=L1; else pc=L11; end
            endcase
	end
    end

endmodule // process
